import logging
import numpy as np
from os.path import join, basename, splitext
from subprocess import call
from argparse import Namespace
import pandas as pd
from typing import Any, Dict, Iterable, List
import pyBigWig

from joblib import Parallel, delayed


def create_bigwigs(matrix, outdir, args):
    # type: (pd.DataFrame, str, Namespace) -> None
    """Create bigwigs from matrix."""
    call("mkdir -p {}".format(outdir), shell=True)
    genome_size_dict = args.chromosome_sizes

    outpaths, data = [], []
    for bed_file in matrix:
        outpath = join(outdir, splitext(basename(bed_file))[0] + ".bw")
        outpaths.append(outpath)

        bed_column = matrix[bed_file]
        data.append(bed_column)

    Parallel(n_jobs=args.number_cores)(delayed(_create_bigwig)(bed_column, outpath, genome_size_dict) for outpath, bed_column in zip(outpaths, data))


def _to_int(l):
    # type: (Iterable[Any]) -> List[int]
    return [int(i) for i in l]


def _create_bigwig(bed_column, outpath, genome_size_dict):
    # type: (pd.Series, str, Dict[str, int]) -> None

    logging.info("Creating biwgwig " + outpath)

    rpkm = 1e6 * bed_column / bed_column.sum()

    rpkm = [float(f) for f in list(rpkm.fillna(0).values.flatten())]

    bed_column = bed_column.reset_index()
    unique_chromosomes = list(bed_column.Chromosome.drop_duplicates())
    chromosomes = list(bed_column.Chromosome)
    starts = _to_int(list(bed_column.Bin))
    ends = _to_int(list(bed_column.End))

    header = [(c, int(genome_size_dict[c])) for c in unique_chromosomes]

    bw = pyBigWig.open(outpath, "w")
    bw.addHeader(header)

    bw.addEntries(chromosomes, starts, ends=ends, values=rpkm)
    bw.close()


def create_sum_bigwigs(matrix, outdir, args):
    # type: (pd.DataFrame, str, Namespace) -> None
    call("mkdir -p {}".format(outdir), shell=True)

    chip = matrix[args.treatment].sum(axis=1)
    input = matrix[args.control].sum(axis=1)

    chip_outpath = join(outdir, "chip_sum" + ".bw")
    input_outpath = join(outdir, "input_sum" + ".bw")

    # _create_bigwig(chip, chip_outpath, args)
    # _create_bigwig(input, input_outpath, args)

    Parallel(n_jobs=args.number_cores)(delayed(_create_bigwig)(bed_column, outpath, args.chromosome_sizes) for outpath, bed_column in zip([chip_outpath, input_outpath], [chip, input]))



import pytest

from collections import namedtuple

import pandas as pd

from epic.config.genomes import create_genome_size_dict

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"

MockNamespace = namedtuple(
    "MockNamespace",
    ["number_cores", "genome", "keep_duplicates", "window_size",
     "fragment_size", "paired_end", "gaps_allowed",
     "false_discovery_rate_cutoff", "effective_genome_size", "store_matrix",
     "bigwig", "bed", "sum_bigwig", "chromosome_sizes",
     "treatment", "control"])

egs = 2290813547.4  # this is the effective genome size used by the original sicer for hg19
gsd = create_genome_size_dict("hg19")


@pytest.fixture(scope="session")
def args_200_fast():
    return MockNamespace(25, "hg19", False, 200, 150, False, 3, 1, egs, False, False, False,
                         False, gsd, ["examples/test.bed"],
                         ["examples/control.bed"])


@pytest.fixture(scope="session")
def args_200():
    return MockNamespace(1, "hg19", False, 200, 150, False, 3, 0.05, egs,
                         False, False, False, False, gsd,
                         ["examples/test.bed"], ["examples/control.bed"])


@pytest.fixture(scope="session")
def args_50():
    return MockNamespace(1, "hg19", False, 50, 150, False, 3, 0.05, egs, False, False, False,
                         False, gsd, ["examples/test.bed"],
                         ["examples/control.bed"])


@pytest.fixture(scope="session")
def expected_result_example_input():
    return pd.read_table("examples/expected_results.csv", sep=" ", skiprows=1)


@pytest.fixture(scope="session")
def epic_overlap_files():
    return ["examples/epic-overlaps/0h_H3K4me3.regions",
"examples/epic-overlaps/3h_H3K4me3.regions",
"examples/epic-overlaps/6h_H3K4me3.regions",
"examples/epic-overlaps/9h_H3K4me3.regions",
"examples/epic-overlaps/12h_H3K4me3.regions",
"examples/epic-overlaps/15h_H3K4me3.regions",
"examples/epic-overlaps/18h_H3K4me3.regions",
"examples/epic-overlaps/21h_H3K4me3.regions",
"examples/epic-overlaps/24h_H3K4me3.regions"]


@pytest.fixture(scope="session")
def epic_overlap_files_more_chromos():
    return ["examples/epic-overlaps/more_chromos0h_H3K4me3.regions",
"examples/epic-overlaps/more_chromos3h_H3K4me3.regions",
"examples/epic-overlaps/more_chromos6h_H3K4me3.regions",
"examples/epic-overlaps/more_chromos9h_H3K4me3.regions",
"examples/epic-overlaps/more_chromos12h_H3K4me3.regions",
"examples/epic-overlaps/more_chromos15h_H3K4me3.regions",
"examples/epic-overlaps/more_chromos18h_H3K4me3.regions",
"examples/epic-overlaps/more_chromos21h_H3K4me3.regions",
"examples/epic-overlaps/more_chromos24h_H3K4me3.regions"]


@pytest.fixture(scope="session")
def epic_overlap_intermediate_nucleotide_matrixes():
    return ["examples/epic-overlaps/0h_H3K4me3.matrix",
            "examples/epic-overlaps/3h_H3K4me3.matrix",
            "examples/epic-overlaps/6h_H3K4me3.matrix"]


@pytest.fixture(scope="session")
def epic_overlap_intermediate_region_matrixes():
    return ["examples/epic-overlaps/0h_H3K4me3_region.matrix",
            "examples/epic-overlaps/3h_H3K4me3_region.matrix",
            "examples/epic-overlaps/6h_H3K4me3_region.matrix"]



import pandas as pd
from numpy import int32
from typing import Any, Tuple, Sequence

def merge_chromosome_dfs(df_tuple):
    # type: (Tuple[pd.DataFrame, pd.DataFrame]) -> pd.DataFrame
    """Merges data from the two strands into strand-agnostic counts."""

    plus_df, minus_df = df_tuple
    index_cols = "Chromosome Bin".split()
    count_column = plus_df.columns[0]

    if plus_df.empty:
        return return_other(minus_df, count_column, index_cols)
    if minus_df.empty:
        return return_other(plus_df, count_column, index_cols)

    # sum duplicate bins
    # TODO: why are there duplicate bins here in the first place?
    plus_df = plus_df.groupby(index_cols).sum()
    minus_df = minus_df.groupby(index_cols).sum()

    # first sum the two bins from each strand
    df = pd.concat([plus_df, minus_df], axis=1).fillna(0).sum(axis=1)
    df = df.reset_index().sort_values(by="Bin")

    df.columns = ["Chromosome", "Bin", count_column]

    df = df.sort_values(["Chromosome", "Bin"])
    df[["Bin", count_column]] = df[["Bin", count_column]].astype(int32)
    df = df[[count_column, "Chromosome", "Bin"]]
    return df.reset_index(drop=True)


def return_other(df, count_column, index_cols):
    # type: (pd.DataFrame, Any, Sequence[Any]) -> pd.DataFrame

    df[[count_column, "Bin"]] = df[[count_column, "Bin"]].astype(int32)
    df = df.groupby(index_cols).sum().reset_index()
    return df[[count_column, "Chromosome", "Bin"]]

# multiprocessing.pool.RemoteTraceback:
# """
# Traceback (most recent call last):
#   File "/local/home/endrebak/anaconda3/lib/python3.5/site-packages/joblib-0.9.4-py3.5.egg/joblib/parallel.py", line 130, in __call__
#     return self.func(*args, **kwargs)
#   File "/local/home/endrebak/anaconda3/lib/python3.5/site-packages/joblib-0.9.4-py3.5.egg/joblib/parallel.py", line 72, in __call__
#     return [func(*args, **kwargs) for func, args, kwargs in self.items]
#   File "/local/home/endrebak/anaconda3/lib/python3.5/site-packages/joblib-0.9.4-py3.5.egg/joblib/parallel.py", line 72, in <listcomp>
#     return [func(*args, **kwargs) for func, args, kwargs in self.items]
#   File "/local/home/endrebak/anaconda3/lib/python3.5/site-packages/bioepic-0.0.9-py3.5.egg/epic/windows/count/merge_chromosome_dfs.py", line 21, in merge_chromosome_dfs
#     df = pd.concat([plus_df, minus_df], axis=1).fillna(0).sum(axis=1)
#   File "/local/home/endrebak/anaconda3/lib/python3.5/site-packages/pandas/tools/merge.py", line 846, in concat
#     return op.get_result()
#   File "/local/home/endrebak/anaconda3/lib/python3.5/site-packages/pandas/tools/merge.py", line 1031, in get_result
#     indexers[ax] = obj_labels.reindex(new_labels)[1]
#   File "/local/home/endrebak/anaconda3/lib/python3.5/site-packages/pandas/indexes/multi.py", line 1422, in reindex
#     raise Exception("cannot handle a non-unique multi-index!")
# Exception: cannot handle a non-unique multi-index!



import logging
from os.path import dirname, join, basename
from subprocess import call
from itertools import chain
from typing import Iterable, Sequence, Tuple
from argparse import Namespace

import pandas as pd

from joblib import Parallel, delayed

from natsort import natsorted

from epic.windows.count.remove_out_of_bounds_bins import remove_bins_with_ends_out_of_bounds
from epic.config.genomes import get_genome_size_file



def write_matrix_files(chip_merged, input_merged, df, args):
    # type: (Dict[str, pd.DataFrame], Dict[str, pd.DataFrame], pd.DataFrame, Namespace) -> None

    matrixes = create_matrixes(chip_merged, input_merged, df, args)

    if args.store_matrix:
        print_matrixes(matrixes, args)

    matrix = pd.concat(matrixes, axis=0)
    matrix = matrix.drop("Enriched", axis=1)
    ends = matrix.index.get_level_values("Bin") + int(args.window_size) - 1
    matrix.insert(0, "End", ends)
    matrix = matrix.set_index("End", append=True)
    matrix = matrix.sort_index(level="Chromosome")

    # TODO: remove out of bounds bins

    if args.bigwig:
        # defer initialization so not run during travis
        from epic.bigwig.create_bigwigs import create_bigwigs
        create_bigwigs(matrix, args.bigwig, args)

    if args.sum_bigwig:
        # defer initialization so not run during travis
        from epic.bigwig.create_bigwigs import create_sum_bigwigs
        create_sum_bigwigs(matrix, args.sum_bigwig, args)


def _create_matrixes(chromosome, chip, input, islands,
                     chromosome_size, window_size):
    # type: (str, Dict[str, pd.DataFrame], Dict[str, pd.DataFrame], pd.DataFrame, int, int) -> pd.DataFrame

    chip_df = get_chromosome_df(chromosome, chip)
    input_df = get_chromosome_df(chromosome, input)

    chip_df["Chromosome"] = chip_df["Chromosome"].astype("category")

    # START workaround
    # Should ideally have been just one line: chip_df["Bin"] = chip_df["Bin"].astype(int)
    # Workaround for the following error:
    # ValueError: assignment destination is read-only
    bins = chip_df["Bin"].astype(int)
    chip_df = chip_df.drop("Bin", axis=1)

    chip_df.insert(0, "Bin", bins)

    # END workaround

    chip_df = chip_df.set_index("Chromosome Bin".split())
    chip_df = islands.join(chip_df, how="right")
    chip_df = chip_df[~chip_df.index.duplicated(keep='first')]


    input_df["Chromosome"] = input_df["Chromosome"].astype("category")

    # START workaround
    # Should ideally have been just one line: input_df["Bin"] = input_df["Bin"].astype(int)
    # Workaround for the following error:
    # ValueError: assignment destination is read-only
    bins = input_df["Bin"].astype(int)
    input_df = input_df.drop("Bin", axis=1)

    input_df.insert(0, "Bin", bins)

    input_df = input_df.set_index("Chromosome Bin".split())

    # END workaround

    input_df = input_df[~input_df.index.duplicated(keep='first')]

    dfm = chip_df.join(input_df, how="outer", sort=False).fillna(0)

    dfm = remove_bins_with_ends_out_of_bounds(dfm, chromosome_size,
                                              window_size)

    return dfm


def create_matrixes(chip, input, df, args):
    # type: (Iterable[pd.DataFrame], Iterable[pd.DataFrame], pd.DataFrame, Namespace) -> List[pd.DataFrame]
    "Creates matrixes which can be written to file as is (matrix) or as bedGraph."

    genome = args.chromosome_sizes

    chip = put_dfs_in_chromosome_dict(chip)
    input = put_dfs_in_chromosome_dict(input)
    all_chromosomes = natsorted(set(list(chip.keys()) + list(input.keys())))

    islands = enriched_bins(df, args)

    logging.info("Creating matrixes from count data.")
    dfms = Parallel(n_jobs=args.number_cores)(delayed(_create_matrixes)(
        chromosome, chip, input, islands, genome[chromosome],
        args.window_size) for chromosome in all_chromosomes)

    return dfms


def print_matrixes(matrixes, args):
    # type: (Iterable[pd.DataFrame], Namespace) -> None
    outpath = args.store_matrix

    dir = dirname(outpath)
    if dir:
        call("mkdir -p {}".format(dir), shell=True)

    logging.info("Writing data matrix to file: " + outpath)
    for i, df in enumerate(matrixes):

        if i == 0:
            header, mode = True, "w+"
        else:
            header, mode = False, "a"

        df.astype(int).to_csv(outpath,
                              sep=" ",
                              na_rep="NA",
                              header=header,
                              mode=mode,
                              compression="gzip",
                              chunksize=1e6)


def get_island_bins(df, window_size, genome, args):
    # type: (pd.DataFrame, int, str, Namespace) -> Dict[str, Set[int]]
    """Finds the enriched bins in a df."""

    # need these chromos because the df might not have islands in all chromos
    chromosomes = natsorted(list(args.chromosome_sizes))

    chromosome_island_bins = {} # type: Dict[str, Set[int]]
    df_copy = df.reset_index(drop=False)
    for chromosome in chromosomes:
        cdf = df_copy.loc[df_copy.Chromosome == chromosome]
        if cdf.empty:
            chromosome_island_bins[chromosome] = set()
        else:
            island_starts_ends = zip(cdf.Start.values.tolist(),
                                     cdf.End.values.tolist())
            island_bins = chain(*[range(
                int(start), int(end), window_size)
                                  for start, end in island_starts_ends])
            chromosome_island_bins[chromosome] = set(island_bins)

    return chromosome_island_bins


def put_dfs_in_dict(dfs):
    # type: (Iterable[pd.DataFrame]) -> Dict[str, pd.DataFrame]
    sample_dict = {}
    for df in dfs:

        if df.empty:
            continue

        chromosome = df.head(1).Chromosome.values[0]
        sample_dict[chromosome] = df

    return sample_dict


def put_dfs_in_chromosome_dict(dfs):
    # type: (Iterable[pd.DataFrame]) -> Dict[str, pd.DataFrame]

    chromosome_dict = {}        # type: Dict[str, pd.DataFrame]
    for df in dfs:

        if df.empty:
            continue

        chromosome = df.head(1).Chromosome.values[0]
        chromosome_dict[chromosome] = df

    return chromosome_dict


def get_chromosome_df(chromosome, df_dict):
    # type: (str, Dict[str, pd.DataFrame]) -> pd.DataFrame

    if chromosome in df_dict:
        df = df_dict[chromosome]
    else:
        df = pd.DataFrame(columns="Chromosome Bin".split())

    return df


def enriched_bins(df, args):
    # type: (pd.DataFrame, Namespace) -> pd.DataFrame

    df = df.loc[df.FDR < args.false_discovery_rate_cutoff]

    idx_rowdicts = []
    for _, row in df.iterrows():
        for bin in range(
                int(row.Start), int(row.End) + 2, int(args.window_size)):
            idx_rowdicts.append({"Chromosome": row.Chromosome,
                                 "Bin": bin,
                                 "Enriched": 1})
    islands = pd.DataFrame.from_dict(idx_rowdicts)
    islands.loc[:, "Chromosome"].astype("category")
    islands.loc[:, "Bin"].astype(int)

    return islands.set_index("Chromosome Bin".split())



from natsort import natsorted
from collections import OrderedDict
import pkg_resources
import logging
from typing import Dict

from epic.config import logging_settings
from epic.utils.find_readlength import (find_readlength,
                                        get_closest_readlength)

__author__ = "Endre Bakken Stovner https://github.com/endrebak/"
__license__ = "MIT"


def get_genome_size_file(genome):
    # type: (str) -> str

    genome_names = pkg_resources.resource_listdir("epic", "scripts/chromsizes")
    name_dict = {n.lower().replace(".chromsizes", ""): n for n in genome_names}

    # No try/except here, because get_egs would already have failed if genome
    # did not exist
    genome_exact = name_dict[genome.lower()]

    return pkg_resources.resource_filename(
        "epic", "scripts/chromsizes/{}".format(genome_exact))


def create_genome_size_dict(genome):
    # type: (str) -> Dict[str,int]
    """Creates genome size dict from string containing data."""

    size_file = get_genome_size_file(genome)
    size_lines = open(size_file).readlines()

    size_dict = {}
    for line in size_lines:
        genome, length = line.split()
        size_dict[genome] = int(length)

    return size_dict


def create_genome_size_dict_custom_genome(chromsizes):
    # type: (str) -> OrderedDict[str, int]

    chromosome_lengths = [l.split() for l in open(chromsizes).readlines()]

    od = OrderedDict()          # type: OrderedDict[str, int]

    for c, l in natsorted(chromosome_lengths):
        od[c] = int(l)

    return od


def get_effective_genome_length(genome, read_length):
    # type: (str, int) -> float

    genome_names = pkg_resources.resource_listdir("epic",
                                                  "scripts/effective_sizes")
    name_dict = {n.split("_")[0]: "".join(n.split("_")[:-1])
                 for n in genome_names}

    try:
        genome_exact = name_dict[genome.lower()]
        egf = pkg_resources.resource_string( # type: ignore
            "epic", "scripts/effective_sizes/{}_{}.txt".format(
                genome_exact, read_length)).split()[-1].decode()
    except KeyError:
        genome_list = "\n".join(list(name_dict.keys()))
        logging.error(
            "Genome " + genome +
            " not found.\n These are the available genomes: " + genome_list +
            "\nIf yours is not there, please request it at github.com/endrebak/epic .")

    genome_length = sum(create_genome_size_dict(genome).values())

    logging.info("Using an effective genome fraction of {}.".format(egf))

    assert float(
        egf) < 1, "Something wrong happened, effective genome fraction over 1!"

    egs = float(egf) * genome_length

    return egs



import pytest

from io import StringIO
import pandas as pd
from numpy import allclose, int32

from epic.windows.count.merge_chromosome_dfs import merge_chromosome_dfs


@pytest.mark.integration
def test_merge_chromosome_dfs(plus_df, minus_df, expected_df):

    actual_df = merge_chromosome_dfs((plus_df, minus_df))
    print(actual_df.to_csv(sep=" "), "actual")
    print(expected_df.to_csv(sep=" "), "expected")
    print(actual_df, "actual")
    print(expected_df, "expected")
    print(actual_df.dtypes, "actual")
    print(expected_df.dtypes, "expected")
    assert actual_df.equals(expected_df)


@pytest.fixture
def plus_df():

    return pd.read_table(
        StringIO(u"""
   Count Chromosome       Bin
2       chr1  39036800
1       chr1  73781000
"""),
        sep=r"\s+")


@pytest.fixture
def minus_df():

    return pd.read_table(
        StringIO(u"""
   Count Chromosome       Bin
5       chr1   39036800
1       chr1   90059600
"""),
        sep=r"\s+")


@pytest.fixture
def expected_df():
    return pd.read_table(
        StringIO(u"""
Count   Chromosome Bin
7      chr1       39036800
1      chr1           73781000
1     chr1        90059600"""),
        sep="\s+",
        dtype={"Count": int32,
               "Bin": int32})



import os
import sys
from setuptools import setup, find_packages
# from Cython.Build import cythonize

version_path = os.path.join(os.path.dirname(__file__), "..", "version.py")
version = open(version_path).readline().split('"')[1]

setup(
    name="bioepic",
    packages=find_packages(),

    # ext_modules=cythonize(
    #      "epic/statistics/add_to_island_expectations_cython.pyx"),
    scripts=["bin/epic", "bin/epic-effective", "bin/epic-overlaps"],
    package_data={'epic': ['scripts/effective_sizes/*.txt',
                           'scripts/chromsizes/*chromsizes',
                           'scripts/genome.snakefile']},
    version=version,
    description="Chip-Seq broad peak/domain finder.",
    author="Endre Bakken Stovner",
    author_email="endrebak85@gmail.com",
    url="http://github.com/endrebak/epic",
    keywords=["ChIP-Seq"],
    license=["MIT"],
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Other Environment", "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Scientific/Engineering"
    ],
    long_description=
    ("Chip-Seq broad peak/domain finder based on SICER. See the url for more info."
     ))



"""This is a module that computes the island threshold.

It is a sanitized and memory efficient version of the SICER version, that
produces results similar to the second or third decimal. I have a less efficient
version that gives even closer results, but I figure that the computations are
very approximate (due to how they are implemented in the original) so I just use
the faster version.
"""

from __future__ import print_function
from argparse import Namespace
from typing import Tuple
import logging

from epic.statistics.compute_values_needed_for_recurrence import (
    compute_enriched_threshold, compute_gap_factor, compute_boundary)
from epic.statistics.compute_score_threshold import compute_score_threshold


# @MEMORY.cache(verbose=0)
def compute_background_probabilities(total_chip_count, args):
    # type: (int, Namespace) -> Tuple[float, int, float]

    effective_genome_size = args.effective_genome_size
    logging.debug(str(effective_genome_size) + " effective_genome_size")
    # move outside of function call

    average_window_readcount = total_chip_count * (
        args.window_size / float(effective_genome_size))
    logging.debug(str(args.window_size) + " window size")
    logging.debug(str(total_chip_count) + " total chip count")
    logging.debug(str(average_window_readcount) + " average_window_readcount")

    island_enriched_threshold = compute_enriched_threshold(
        average_window_readcount)
    logging.debug(str(island_enriched_threshold) +
                  " island_enriched_threshold")

    gap_contribution = compute_gap_factor(
        island_enriched_threshold, args.gaps_allowed, average_window_readcount)
    logging.debug(str(gap_contribution) + " gap_contribution")

    boundary_contribution = compute_boundary(
        island_enriched_threshold, args.gaps_allowed, average_window_readcount)
    logging.debug(str(boundary_contribution) + " boundary_contribution")

    genome_length_in_bins = effective_genome_size / args.window_size

    score_threshold = compute_score_threshold(
        average_window_readcount, island_enriched_threshold, gap_contribution,
        boundary_contribution, genome_length_in_bins)

    return score_threshold, island_enriched_threshold, average_window_readcount
